"""Unit tests for the rule registry."""

import pytest
from dbt_score import Severity
from dbt_score.config import Config
from dbt_score.exceptions import DuplicatedRuleException
from dbt_score.rule_registry import RuleRegistry


def test_rule_registry_discovery(default_config):
    """Ensure rules can be found in a given namespace recursively."""
    r = RuleRegistry(default_config)
    r._load("tests.rules")
    assert sorted(r._rules.keys()) == [
        "tests.rules.nested.example.rule_test_nested_example",
        "tests.rules.rules.rule_test_example",
    ]
    assert list(r._rule_filters.keys()) == [
        "tests.rules.rule_filters.skip_model1",
        "tests.rules.rule_filters.skip_schemaX",
    ]


def test_disabled_rule_registry_discovery():
    """Ensure disabled rules are not discovered."""
    config = Config()
    config.disabled_rules = ["tests.rules.nested.example.rule_test_nested_example"]
    r = RuleRegistry(config)
    r._load("tests.rules")
    assert sorted(r._rules.keys()) == [
        "tests.rules.rules.rule_test_example",
    ]


def test_configured_rule_registry_discovery(valid_config_path):
    """Ensure rules are discovered and configured correctly."""
    config = Config()
    config._load_toml_file(str(valid_config_path))
    r = RuleRegistry(config)
    r._load("tests.rules")
    assert r.rules["tests.rules.rules.rule_test_example"].severity == Severity.CRITICAL


def test_rule_registry_no_duplicates(default_config):
    """Ensure no duplicate rule names can coexist."""
    r = RuleRegistry(default_config)
    r._load("tests.rules")
    with pytest.raises(DuplicatedRuleException):
        r._load("tests.rules")


def test_rule_registry_core_rules(default_config):
    """Ensure core rules are automatically discovered."""
    r = RuleRegistry(default_config)
    r.load_all()
    assert len(r.rules) > 0


def test_rule_registry_rule_filters(valid_config_path, model1, model2):
    """Test config filters are loaded."""
    config = Config()
    config._load_toml_file(str(valid_config_path))
    r = RuleRegistry(config)
    r._load("tests.rules")
    r._load_filters_into_rules()

    assert not r.rules["tests.rules.rules.rule_test_example"].should_evaluate(model1)
    assert r.rules["tests.rules.rules.rule_test_example"].should_evaluate(model2)



"""Test models."""

from pathlib import Path
from unittest.mock import patch

from dbt_score.models import ManifestLoader


@patch("dbt_score.models.Path.read_text")
def test_manifest_load(mock_read_text, raw_manifest):
    """Test loading a manifest."""
    with patch("dbt_score.models.json.loads", return_value=raw_manifest):
        loader = ManifestLoader(Path("some.json"))
        assert len(loader.models) == len(
            [
                node
                for node in raw_manifest["nodes"].values()
                if node["resource_type"] == "model"
                and node["package_name"] == raw_manifest["metadata"]["project_name"]
            ]
        )

        model1 = loader.models["model.package.model1"]
        assert model1.tests[0].name == "test2"
        assert model1.tests[1].name == "test4"
        assert model1.columns[0].tests[0].name == "test1"

        assert len(loader.sources) == len(
            [
                source
                for source in raw_manifest["sources"].values()
                if source["package_name"] == raw_manifest["metadata"]["project_name"]
            ]
        )
        assert (
            loader.sources["source.package.my_source.table1"].tests[0].name
            == "source_test1"
        )

        assert loader.snapshots["snapshot.package.snapshot1"].parents == [
            loader.models["model.package.model1"]
        ]
        assert loader.models["model.package.model1"].children == [
            loader.snapshots["snapshot.package.snapshot1"]
        ]
        assert loader.models["model.package.model1"].parents == [
            loader.models["model.package.model2"],
            loader.sources["source.package.my_source.table1"],
            loader.snapshots["snapshot.package.snapshot2"],
        ]
        assert loader.models["model.package.model2"].children == [
            loader.models["model.package.model1"]
        ]
        assert loader.models["model.package.model2"].parents == [
            loader.seeds["seed.package.seed1"]
        ]
        assert loader.snapshots["snapshot.package.snapshot2"].parents == [
            loader.sources["source.package.my_source.table1"]
        ]
        assert loader.sources["source.package.my_source.table1"].children == [
            loader.models["model.package.model1"],
            loader.snapshots["snapshot.package.snapshot2"],
        ]
        assert loader.seeds["seed.package.seed1"].children == [
            loader.models["model.package.model2"]
        ]


@patch("dbt_score.models.Path.read_text")
def test_manifest_select_models_simple(mock_read_text, raw_manifest):
    """Test a simple selection in a manifest."""
    with patch("dbt_score.models.json.loads", return_value=raw_manifest):
        manifest_loader = ManifestLoader(Path("some.json"), select=["model1"])

    assert [x.name for x in manifest_loader.models.values()] == ["model1"]


@patch("dbt_score.models.Path.read_text")
@patch("dbt_score.models.dbt_ls")
def test_manifest_select_models_dbt_ls(mock_dbt_ls, mock_read_text, raw_manifest):
    """Test a complex selection in a manifest, which uses dbt ls."""
    mock_dbt_ls.return_value = ["model1"]
    with patch("dbt_score.models.json.loads", return_value=raw_manifest):
        manifest_loader = ManifestLoader(Path("some.json"), select=["+model1"])

    assert [x.name for x in manifest_loader.models.values()] == ["model1"]
    mock_dbt_ls.assert_called_once_with(["+model1"])


@patch("dbt_score.models.Path.read_text")
@patch("dbt_score.models.dbt_ls")
def test_manifest_no_model(mock_dbt_ls, mock_read_text, raw_manifest, caplog):
    """Test the lack of model in a manifest."""
    with patch("dbt_score.models.json.loads", return_value=raw_manifest):
        manifest_loader = ManifestLoader(Path("some.json"), select=["non_existing"])

    assert len(manifest_loader.models) == 0
    assert "Nothing to evaluate!" in caplog.text
